

import numpy as np
import time

from .runningstats import call_roi_stats


class runningstats():

    def __init__(self, verbose=False):
        
        self.verbose = verbose
        self.module = dict()
        self.dtype = np.float32


    def update_roi(self, list_of_module_name_and_online_roi_pairs):
        for pair in list_of_module_name_and_online_roi_pairs:
            module_name = pair[0]

            self.ensure_init_module(module_name)
            online_roi = pair[1]

            self.module[module_name].update(online_roi)

    def process_detector_data_list(self,
        detector_data_list, # list of pairs (unique detecot panel name, detector data of shape <numbunches, slow scan dim, fast scan dim>)
        lit_frames = None, # if lit frames is None, then all frames are processed, otherwise selects 
        # lit frames is a vector of ints that all are in the interval [0, numbunches) 

        # returns a list of dicts with entries
        # "module_name"
        # "bunch_resolved_roi__stats"
        # "pixel_stats"
        # "processing_time"
    ):
        
        if lit_frames is None:
            lit_frames = -1 * np.ones((1,))
        lit_frames = lit_frames.astype(np.int16)
        
        result = list()

        for pair in detector_data_list:
            module_name = pair[0]

            if module_name not in self.module:
                print(f"runningstats Error: {module_name} is not initialized! You did probably not call update roi with this module name and a corresponding online roi")
                continue
            
            module_data = pair[1]

            if module_data.ndim != 3:
                print(f"Error in Stats: data shape is not equal to 3, but {module_data.ndim} and of shape {module_data.shape}")
                continue
            
            num_images = module_data.shape[0]
            image_size = module_data.shape[1] * module_data.shape[2]
            
            # check size of num bunches vector, cant be larger than int16 allows (because this is the type for lit frame indices)
            if module_data.shape[0] > 32765:
                print("Error: in process_detector_data_list in runningstats for module {module_name}: no support for module_data.shape[0], i.e., number of bunches, to be {module_data.shape[0]}. max value is '32765'")
                continue

            if (image_size != self.module[module_name]['roi_datasize']):
                print(f"Error: in process_detector_data_list in runningstats for module {module_name}: roi has image size size {self.module[module_name]['roi_datasize']} and provided module data has size {image_size}! they are not equal, aborting")
                continue
            
            bunch_resolved_roi__stats, pixel_stats, time_to_process = self.process(
                module_data,
                num_images,
                image_size, 
                lit_frames, 
                self.module[module_name]['online_roi'])
            
            result.append({
                "module_name", module_name, 
                "bunch_resolved_roi__stats", bunch_resolved_roi__stats,
                "pixel_stats", pixel_stats,
                "processing_time", time_to_process,
                })
        
        return result
            
    
    def ensure_init_module(self, module_name):
        if module_name not in self.module:
            if self.verbose == True:
                print(f"runningstats: initialized module {module_name}")
            self.module[module_name] = dict()
        

    def process(self,
            module_data, 
            num_images,
            image_size,
            lit_frames,
            online_roi):
        '''
        returns bunch_resolved_roi_stats, 
        pixel_stats
        '''
        
        if module_data.dtype != self.dtype:
            print(f"Stats warning: data expected to be {self.dtype}")
            print("data is of type", module_data.dtype)
            print("casting to", self.dtype)
            module_data = module_data.astype(self.dtype)
        
        num_rois = online_roi.shape[0]

        def get_partial_stats_memory(num_images, num_rois):
            return np.zeros((num_images, num_rois,3), np.float64)


        memory_for_bunch_resolved_roi_stats = get_partial_stats_memory(num_images, num_rois)
        memory_for_pixel_stats = get_partial_stats_memory(image_size,1)

        time_to_call = time.perf_counter()

        num_rois = online_roi.shape[0]
        online_roi_max_length = online_roi.shape[1]
        call_roi_stats(
            module_data.reshape(-1), # data
            image_size, # size of one image inside data
            num_images, # number of images in the data
            online_roi, # roi 
            num_rois, # number of rois
            online_roi_max_length, # max length of compressed roi
            lit_frames.reshape(-1),
            lit_frames.size,
            memory_for_bunch_resolved_roi_stats.reshape(-1), # memory 
            memory_for_pixel_stats.reshape(-1) #memory to bynch
            ) 
        
        time_to_call = time.perf_counter() - time_to_call
        return memory_for_bunch_resolved_roi_stats, memory_for_pixel_stats, time_to_call
####
        


            


    



        
