
#include <array>
#include <chrono>
#include <cmath>
#include <cstdint>
#include <experimental/simd>
#include <memory>
#include <iostream>
#include <optional>
#include <span>
#include <type_traits>
#include <utility>
#include <vector>

#include "task_thread_pool.hpp"
#include "runningstats_impl.h"






struct running_mean
{
    std::uint64_t count{0};
    double mean{0.0};

    void add(double value)
    {
        count++;
        mean += (value - mean) / static_cast<double>(count);    
    }

};

namespace stdx = std::experimental;

/*
struct running_mean_simd
{
    stdx::native_simd<float> mean{0};
    stdx::native_simd<float> count{0};

    void add(float* data_ptr)
    {

        //count++;
        //mean += (value - mean) / static_cast<double>(count);    




        stdx::native_simd<float> nan_sum_simd{0};
    stdx::native_simd<float> num_values_added{0};

    
    }

};*/


void running_stats(
    float* data,
    std::uint32_t data_size,
    double* running_mean_mem, // of size data_size*2
    float* nanmean_of_data) // of size   1
{
    // necessary because the (double* running_mean_mem) input serves as storage for running_mean stucts
    static_assert(2*sizeof(double)==sizeof(running_mean));

    // initialize nan mean that is used to compute nan mean for this data set
    running_mean nan_mean;

    // run through data 
    for (std::uint32_t idx = 0; idx < data_size; idx++)
    {   
        const double value = static_cast<double>(data[idx]);

        // we do not process non finite values
        if(!std::isfinite(value))
        {
            continue;
        }
        
        // add to nan mean
        nan_mean.add(value);

        // add to mean 
        reinterpret_cast<running_mean*>(running_mean_mem+2*idx)->add(value);
    }

    *nanmean_of_data = nan_mean.mean;

    return;
}



 
void println(std::string_view name, auto const& a)
{
    std::cout << name << ": ";
    for (std::size_t i{}; i != std::size(a); ++i)
        std::cout << a[i] << ' ';
    std::cout << '\n';
}

bool is_aligned(const void *pointer, size_t byte_count)
{ 
    //std::cout << "is aligned " << (std::uintptr_t)pointer % byte_count << "\n";
    return (std::uintptr_t)pointer % byte_count == 0; 
}

inline bool is_data_aligned_for_simd(const double* incoming_data_ptr, const double* running_mean_mem_ptr)
{
    // verify memory compatability with running_mean_struct
    // necessary because the (double* running_mean_mem) input serves as storage for running_mean stucts
    static_assert(2*sizeof(double)==sizeof(running_mean));

    constexpr auto align_val_data = stdx::memory_alignment_v<stdx::native_simd<double>>;
    const bool data_is_aligned = is_aligned(incoming_data_ptr,align_val_data);
    //std::cout << "DOUBLE: data_is_aligned? " << data_is_aligned << " &data " << incoming_data_ptr << "\n";
    
    constexpr auto align_val_mem = stdx::memory_alignment_v<stdx::native_simd<double>>;
    static_assert(align_val_mem == stdx::memory_alignment_v<stdx::native_simd<std::uint64_t>>);
    const bool mem_is_aligned = is_aligned(running_mean_mem_ptr,align_val_mem);
    //std::cout << "DOUBLE: mem_is_aligned? " << mem_is_aligned << " &running_mean_mem_ptr " << running_mean_mem_ptr << "\n";
    
    return (mem_is_aligned && data_is_aligned);
}

inline bool is_data_aligned_for_simd_float(const float* incoming_data_ptr, const float* running_mean_mem_ptr)
{
    // verify memory compatability with running_mean_struct
    // necessary because the (double* running_mean_mem) input serves as storage for running_mean stucts
    //static_assert(2*sizeof(double)==sizeof(running_mean));

    constexpr auto align_val_data = stdx::memory_alignment_v<stdx::native_simd<float>>;
    const bool data_is_aligned = is_aligned(incoming_data_ptr,align_val_data);
    //std::cout << "FLOAT: data_is_aligned? " << data_is_aligned << " &data " << incoming_data_ptr << "\n";
    
    constexpr auto align_val_mem = stdx::memory_alignment_v<stdx::native_simd<float>>;
    static_assert(align_val_mem == stdx::memory_alignment_v<stdx::native_simd<std::uint32_t>>);
    const bool mem_is_aligned = is_aligned(running_mean_mem_ptr,align_val_mem);
    //std::cout << "FLOAT: mem_is_aligned? " << mem_is_aligned << " &running_mean_mem_ptr " << running_mean_mem_ptr << "\n";
    
    return (mem_is_aligned && data_is_aligned);
}


    
void running_stats_simd(
    double* data,
    std::uint32_t data_size,
    double* running_mean_mem, // of size data_size*2
    std::uint32_t* simd_offset,
    float* nanmean_of_data,
    std::uint32_t* vec_size,
    std::uint32_t numThreads) // of size   1
{
    
    const auto offset = ((std::uintptr_t)running_mean_mem % stdx::memory_alignment_v<stdx::native_simd<double>>)/sizeof(double);
    *simd_offset = offset;
    running_mean_mem += offset;

    if (!is_data_aligned_for_simd(data, running_mean_mem))
    {
        std::cout << "ERROR, MEMORY NOT ALIGNED for simd vector_aligned!\n";
        return;
    }

    constexpr auto align_val_data = stdx::memory_alignment_v<stdx::native_simd<double>>;
    constexpr auto num_elements_per_data_vec = align_val_data/sizeof(double);
    *vec_size = num_elements_per_data_vec;
    //const std::size_t numThreads{numThreads};
    static task_thread_pool::task_thread_pool pool{numThreads};
    const auto data_size_mod_vec_size = data_size % (numThreads * num_elements_per_data_vec);
    if (data_size_mod_vec_size != 0)
    {
        std::cout << "Error, data size does not fit threads and simd vec size (threads = " << numThreads << " ), (simd vec size = " << num_elements_per_data_vec << " ) , (data size = " << data_size << " )\n";
        return;
    }
    std::uint32_t size_per_thread = data_size/numThreads;
    //std::cout << "size per thread " << size_per_thread << "\n"; 
    auto process_stat_simd = [size_per_thread](double* data, double* running_mean_mem, float* nansum, float* nannum) -> void
    {
        //std::cout << "data IT " << data << "   running_mean_mem " << running_mean_mem << "   nanmean_of_data " << nanmean_of_data_ << "\n";
        constexpr auto align_val_data = stdx::memory_alignment_v<stdx::native_simd<double>>;
        constexpr auto num_elements_per_data_vec = align_val_data/sizeof(double);
        stdx::native_simd<double> nan_sum_simd{0};
        stdx::native_simd<double> num_values_added{0};
        stdx::native_simd<double> data_vec;// uninitialized
        stdx::native_simd<double> running_vec;// uninitialized
        stdx::native_simd<double> running_vec_counter;// uninitialized
        static_assert(num_elements_per_data_vec==running_vec.size());
        static_assert(num_elements_per_data_vec==running_vec_counter.size());

        double* memory = running_mean_mem;
        for (std::uint32_t idx = 0; idx<size_per_thread; idx += num_elements_per_data_vec)
        {
        // read values
        //data_vec.copy_from(&data[idx], stdx::vector_aligned);
        data_vec.copy_from(&data[idx], stdx::vector_aligned);
        //println("data vec after copy", data_vec);
        running_vec.copy_from(memory+2*idx, stdx::vector_aligned);
        running_vec_counter.copy_from(memory+2*idx+num_elements_per_data_vec, stdx::vector_aligned);

        // only select finite
        const auto finite = stdx::isfinite(data_vec);
        stdx::where(!finite, data_vec) = 0;
        // add nan mean
        nan_sum_simd += data_vec;
        running_vec += data_vec;
        stdx::where(finite, data_vec) = 1;
        num_values_added += data_vec;
        // add nan means
        running_vec_counter += data_vec;
        running_vec.copy_to(memory+2*idx, stdx::vector_aligned);
        running_vec_counter.copy_to(memory+2*idx+num_elements_per_data_vec, stdx::vector_aligned);        
        }
        *nansum = static_cast<float>(stdx::reduce(nan_sum_simd,std::plus{}));
        *nannum = static_cast<float>(stdx::reduce(num_values_added,std::plus{}));
        return;
    };

    std::vector<float> nansum(numThreads);
    std::vector<float> nannum(numThreads);
    for (std::uint32_t numTask = 0; numTask < numThreads; ++numTask)
    {
        pool.submit_detach(process_stat_simd, data+numTask*size_per_thread, running_mean_mem+numTask*2*size_per_thread, nansum.data()+numTask, nannum.data()+numTask);
    }
    pool.wait_for_tasks();
    float nan_mean = 0.0;
    for (auto dat : nansum)
    {
        nan_mean += dat;
    }
    float nannum_all = 0.0;
    for (auto dat : nannum)
    {
        nannum_all += dat;
    }
    if (nannum_all == 0.0f)
    {
        *nanmean_of_data = INFINITY;
    }
    else
    {
        *nanmean_of_data = nan_mean/nannum_all;
    }

    return;
}



void running_stats_simd_float(
    float* data,
    std::uint32_t data_size,
    float* running_mean_mem, // of size data_size*2
    std::uint32_t* simd_offset,
    float* nanmean_of_data,
    std::uint32_t* vec_size,
    std::uint32_t numThreads) // of size   1
{

    const auto offset = ((std::uintptr_t)running_mean_mem % stdx::memory_alignment_v<stdx::native_simd<float>>)/sizeof(float);
    *simd_offset = offset;
    running_mean_mem += offset;

    if (!is_data_aligned_for_simd_float(data, running_mean_mem))
    {
        std::cout << "ERROR, MEMORY NOT ALIGNED for simd float vector_aligned!\n";
        return;
    }

    constexpr auto align_val_data = stdx::memory_alignment_v<stdx::native_simd<float>>;
    constexpr auto num_elements_per_data_vec = align_val_data/sizeof(float);
    *vec_size = num_elements_per_data_vec;
    //const std::size_t numThreads{numThreads};
    static task_thread_pool::task_thread_pool pool{numThreads};
    const auto data_size_mod_vec_size = data_size % (numThreads * num_elements_per_data_vec);
    if (data_size_mod_vec_size != 0)
    {
        std::cout << "Error, data size does not fit threads and simd vec size (threads = " << numThreads << " ), (simd vec size = " << num_elements_per_data_vec << " ) , (data size = " << data_size << " )\n";
        return;
    }
    std::uint32_t size_per_thread = data_size/numThreads;
    //std::cout << "size per thread " << size_per_thread << "\n"; 
    auto process_stat_simd = [size_per_thread](float* data, float* running_mean_mem, float* nansum, float* nannum) -> void
    {
        //std::cout << "data IT " << data << "   running_mean_mem " << running_mean_mem << "   nanmean_of_data " << nanmean_of_data_ << "\n";
        constexpr auto align_val_data = stdx::memory_alignment_v<stdx::native_simd<float>>;
        constexpr auto num_elements_per_data_vec = align_val_data/sizeof(float);
        stdx::native_simd<float> nan_sum_simd{0};
        stdx::native_simd<float> num_values_added{0};
        stdx::native_simd<float> data_vec;// uninitialized
        stdx::native_simd<float> running_vec;// uninitialized
        stdx::native_simd<float> running_vec_counter;// uninitialized
        static_assert(num_elements_per_data_vec==running_vec.size());
        static_assert(num_elements_per_data_vec==running_vec_counter.size());

        float* memory = running_mean_mem;
        for (std::uint32_t idx = 0; idx<size_per_thread; idx += num_elements_per_data_vec)
        {
        // read values
        //data_vec.copy_from(&data[idx], stdx::vector_aligned);
        data_vec.copy_from(&data[idx], stdx::vector_aligned);
        //println("data vec after copy", data_vec);
        running_vec.copy_from(memory+2*idx, stdx::vector_aligned);
        running_vec_counter.copy_from(memory+2*idx+num_elements_per_data_vec, stdx::vector_aligned);

        // only select finite
        const auto finite = stdx::isfinite(data_vec);
        stdx::where(!finite, data_vec) = 0;
        // add nan mean
        nan_sum_simd += data_vec;
        running_vec += data_vec;
        stdx::where(finite, data_vec) = 1;
        num_values_added += data_vec;
        // add nan means
        running_vec_counter += data_vec;
        running_vec.copy_to(memory+2*idx, stdx::vector_aligned);
        running_vec_counter.copy_to(memory+2*idx+num_elements_per_data_vec, stdx::vector_aligned);        
        }
        *nansum = stdx::reduce(nan_sum_simd,std::plus{});
        *nannum = stdx::reduce(num_values_added,std::plus{});
        return;
    };

    std::vector<float> nansum(numThreads);
    std::vector<float> nannum(numThreads);
    for (std::uint32_t numTask = 0; numTask < numThreads; ++numTask)
    {
        pool.submit_detach(process_stat_simd, data+numTask*size_per_thread, running_mean_mem+numTask*2*size_per_thread, nansum.data()+numTask, nannum.data()+numTask);
    }
    pool.wait_for_tasks();
    float nan_mean = 0.0;
    for (auto dat : nansum)
    {
        nan_mean += dat;
    }
    float nannum_all = 0.0;
    for (auto dat : nannum)
    {
        nannum_all += dat;
    }
    if (nannum_all == 0.0f)
    {
        *nanmean_of_data = INFINITY;
    }
    else
    {
        *nanmean_of_data = nan_mean/nannum_all;
    }

    return;
}






struct RunningStats
{
    struct Stats
    {
        std::uint32_t numValues;
        double mean;
        double variance;
    };

    std::uint32_t numValues{0};
    double mean{0.0};
    double m2{0.0};
    
    void add(float value)
    {
        const bool isFinite = std::isfinite(value);
        if (!isFinite)
        {
            return;
        }
        auto dvalue = static_cast<double>(value);
        numValues += 1;
        const auto delta = dvalue - mean;
        mean += delta / static_cast<double>(numValues);
        m2 += delta*(dvalue - mean);
    }

    void add(std::optional<float> value)
    {
        if (!value)
        {
            return;
        }
        const float val = value.value();
        numValues += 1;
        const auto delta = val - mean;
        mean += delta / static_cast<float>(numValues);
        m2 += delta*(val - mean);
    }

    std::optional<Stats> get() const
    {
        if (numValues < 1)
        {
            return std::nullopt;
        }

        //const float numValues_inv = 1 / static_cast<float>(numValues-1);
        const double numValues_inv = 1 / static_cast<double>(numValues);
        const double variance = m2 * numValues_inv;
        
        return Stats({numValues, mean, variance});
    }
};





// transpose:

std::shared_ptr<float[]>  tranpose(
    task_thread_pool::task_thread_pool &pool, 
    const float* imgData,  
    std::uint32_t slow_dim_in, 
    std::uint32_t fast_dim_in)
{
    const std::uint32_t fastBlockSizeDim = 2048;   // 2048; // 1024; // can be changed
    const std::uint32_t slowBlockSizeDim = 8; // can be changed
    const std::size_t fastDim = fast_dim_in;
    const std::size_t slowDim = slow_dim_in;
    std::shared_ptr<float[]> transposed_mem = make_shared_for_overwrite<float[]>(fast_dim_in*slow_dim_in);
    float *out_ptr = transposed_mem.get();
    // add regular patches
    std::size_t fastBlockPos{0};
    std::size_t slowBlockPos{0};
    std::size_t fastBlockNum{0};
    std::size_t slowBlockNum{0};

    auto tranpose_block = [imgData, out_ptr, slowDim, fastDim](std::size_t slowBlockStart, std::size_t slowBlockSize, std::size_t fastBlockStart, std::size_t fastBlockSize)
    {
        for (std::size_t this_fast = fastBlockStart; this_fast < fastBlockStart+fastBlockSize; this_fast++)
        {
            for (std::size_t this_slow = slowBlockStart; this_slow < slowBlockStart+slowBlockSize; this_slow++)
            {
                out_ptr[this_fast*slowDim + this_slow] = imgData[this_slow*fastDim + this_fast];
            }
        }
    };

    // t1 = high_resolution_clock::now();
    // auto sparse_after_transpose = std::unique_ptr<std::uint16_t[]>(new std::uint16_t[32*numPhotonsCount]);
    if (slowDim >= slowBlockSizeDim && fastDim >= fastBlockSizeDim)
    {
        while (slowBlockPos + slowBlockSizeDim <= slowDim)
        {
            fastBlockPos = 0;
            fastBlockNum = 0;
            while (fastBlockPos + fastBlockSizeDim <= fastDim)
            {
                pool.submit_detach(
                    tranpose_block, slowBlockPos, slowBlockSizeDim, fastBlockPos, fastBlockSizeDim
                );
                
                fastBlockPos += fastBlockSizeDim;
                fastBlockNum += 1;
            }

            slowBlockPos += slowBlockSizeDim;
            slowBlockNum += 1;
        }
    }

    const std::size_t fastEnd = fastBlockPos;
    const std::size_t slowEnd = slowBlockPos;

    const std::size_t fastLeftover = fastDim - fastEnd;
    const std::size_t slowLeftover = slowDim - slowEnd;

    // check for leftovers
    if (fastLeftover != 0 && slowDim >= slowBlockSizeDim)
    {
        slowBlockPos = 0;
        slowBlockNum = 0;
        while (slowBlockPos + slowBlockSizeDim <= slowDim)
        {
            pool.submit_detach(
                tranpose_block, slowBlockPos, slowBlockSizeDim, fastEnd, fastLeftover);

            slowBlockPos += slowBlockSizeDim;
            slowBlockNum += 1;
        }

        slowBlockPos += slowBlockSizeDim;
        slowBlockNum += 1;
    }

    // check for leftovers
    if (slowLeftover != 0 && fastDim >= fastBlockSizeDim)
    {
        fastBlockPos = 0;
        fastBlockNum = 0;
        while (fastBlockPos + fastBlockSizeDim < fastDim)
        {
            pool.submit_detach(
                tranpose_block, slowEnd, slowLeftover, fastBlockPos, fastBlockSizeDim);

            fastBlockPos += fastBlockSizeDim;
            fastBlockNum += 1;
        }

        fastBlockPos += fastBlockSizeDim;
        fastBlockNum += 1;
    }

    if (slowLeftover != 0 && fastLeftover != 0)
    {
        
        pool.submit_detach(
            tranpose_block, slowEnd, slowLeftover, fastEnd, fastLeftover);
    }

    pool.wait_for_tasks();

#if PRINT
    t2 = high_resolution_clock::now();
    auto us_int_transposing = static_cast<std::int64_t>(duration_cast<microseconds>(t2 - t1).count());
    std::cout << "transposing and photonizing us:" << us_int_transposing << "\n";

    t1 = high_resolution_clock::now();
#endif

    return transposed_mem;
}



void stats_with_roi(
    float* data, // of shape num_images x image_size
    std::uint32_t image_size,
    std::uint32_t num_images,
    std::uint32_t* roi, //  of shape  num_rois x online_roi_length_per_roi
    std::uint32_t num_rois,
    std::uint32_t online_roi_length_per_roi,
    std::int16_t* litframes,
    std::uint16_t num_litframes,
    double* memory_mean, // roi memory of shape   num_images * num_rois * 3
    double* image_memory_mean) // memory of shape   image_size * 3
{
    static task_thread_pool::task_thread_pool pool{32};

    // lambda that procsses an image of image size
    auto process_image = [online_roi_length_per_roi](const float* image_data, const std::uint32_t* compressed_roi, double* this_image_memory_mean)
    {
        RunningStats local_roi_stats;

        for (std::size_t compressed_roi_read_pos = 0; compressed_roi_read_pos<online_roi_length_per_roi; compressed_roi_read_pos+=2)
        {
            const auto image_data_idx = compressed_roi[compressed_roi_read_pos];
            const auto length = compressed_roi[compressed_roi_read_pos+1];

            if (length == 0)
            {
                // finished this roi
                break;
            }
            
            for(std::size_t idx = image_data_idx; idx < image_data_idx+length ; idx++)
            {
                // get the pixel value at this idx
                const float value = image_data[idx];

                // add pixel value to stats // note, non finite value is check inside RunningStats class
                local_roi_stats.add(value);
            }
        }

        // write result into memory
        const auto res = local_roi_stats[this_roi].get();
        if (!res)
        {
            this_image_memory_mean[0] = 0.0;
            this_image_memory_mean[1] = INFINITY;
            this_image_memory_mean[2] = INFINITY;
        }
        else
        {
            // write into buffer
            this_image_memory_mean[0] = static_cast<double>(res.value().numValues);
            this_image_memory_mean[1] = static_cast<double>(res.value().mean);
            this_image_memory_mean[2] = static_cast<double>(res.value().variance);
        }       
    };

    auto call_process_image_num = [&](const std::size_t image_num, const std::size_t roi_num)
    {
        // handle image num
        const float* this_image_data = data + image_num*image_size;
        const std::uint32_t* this_roi = roi+roi_num*online_roi_length_per_roi;
        double* this_image_memory_mean = memory_mean + 3*num_rois*image_num + 3*roi_num;
        
        pool.submit_detach(
            process_image, 
                this_image_data,
                this_roi, 
                this_image_memory_mean);
    };

    if (litframes[0] == -1)
    { 
    // run through all rois
    for(std::size_t roi_num = 0; roi_num < num_rois; roi_num++)
    {
        // only if this roi has pixel in this data we compute!
        if (roi[roi_num*online_roi_length_per_roi+1] != 0) 
        {            
            // process all images in the data set for this roi:      
            for (std::size_t image_num = 0; image_num<num_images; image_num++)
            {
                call_process_image_num(image_num, roi_num);
            }
        }
    }
    }
    else
    { 
    // run through all rois
    for(std::size_t roi_num = 0; roi_num < num_rois; roi_num++)
    {
        // only if this roi has pixel in this data we compute!
        if (roi[roi_num*online_roi_length_per_roi+1] != 0) 
        {
            // process litframes images in the data set    
            for (std::uint32_t litframe_idx = 0; litframe_idx < num_litframes; litframe_idx++ )
            {
                const auto image_num = litframes[litframe_idx];
                call_process_image_num(image_num);        
            }
        }
    }
    }

    auto transposed_data = tranpose(
        pool, 
        data,  
        num_images, 
        image_size);

    pool.wait_for_tasks();

    const std::uint32_t numPixelPerTask = 1024;
    //const std::uint32_t numPixelPerTask = 128;
    if (image_size % numPixelPerTask != 0)
    {
        std::cout << "WARNING FROM STAST WITH ROI: (image_size % numPixelPerTask != 0) " << image_size << " % " << numPixelPerTask << " = " << image_size % numPixelPerTask <<"\n";
    }
    // need to tranverse pixel by pixel but only, lit frames.
    auto process_litframes_pixel = [numPixelPerTask, litframes,num_litframes, num_images](const float* pixelMemStart, double* thisPixelOutputMem)
    {
           
        for (std::uint32_t pixelNum=0; pixelNum < numPixelPerTask; pixelNum++)
        {
            RunningStats local_stats;
            const auto ptr = pixelMemStart+num_images*pixelNum;
            
        for (std::uint32_t litframeIdx = 0; litframeIdx<num_litframes; litframeIdx++)
        {
            const auto image_num = litframes[litframeIdx];
            const float value = ptr[image_num];
        
            // add pixel to roi  
            // note, non finite check is done insise RunningStats
            local_stats.add(value);
        }

        // get the stats
        const auto res = local_stats.get();
        const auto outputPtr = thisPixelOutputMem + 3*pixelNum;
        if (!res)
        {
            outputPtr[0] = 0.0;
            outputPtr[1] = INFINITY;
            outputPtr[2] = INFINITY;
        }
        else
        {
            outputPtr[0] = static_cast<double>(res.value().numValues);
            outputPtr[1] = static_cast<double>(res.value().mean);
            outputPtr[2] = static_cast<double>(res.value().variance);       
        }
        }
    };

    auto process_pixel = [numPixelPerTask, num_images](const float* pixelMemStart, double* thisPixelOutputMem)
    {
        for (std::uint32_t pixelNum=0; pixelNum < numPixelPerTask; pixelNum++)
        {
            RunningStats local_stats;
            const auto ptr = pixelMemStart+num_images*pixelNum;
            
            for (std::uint32_t image_num = 0; image_num<num_images; image_num++)
            {
                const float value = ptr[image_num];
                // add pixel to roi 
                local_stats.add(value);
            }

            // get the stats
            const auto res = local_stats.get();
            const auto outputPtr = thisPixelOutputMem + 3*pixelNum;
            if (!res)
            {
                outputPtr[0] = 0.0
                outputPtr[1] = INFINITY;
                outputPtr[2] = INFINITY;
            }
            else
            {
                outputPtr[0] = static_cast<double>(res.value().numValues);
                outputPtr[1] = static_cast<double>(res.value().mean);
                outputPtr[2] = static_cast<double>(res.value().variance);    
            }
        }   
    };
    
    if (litframes[0] == -1)
    { // process all images in the data set    
       
    for (std::uint32_t this_pixel = 0; this_pixel<image_size; this_pixel+=numPixelPerTask)
    {
        pool.submit_detach(
            process_pixel,
            transposed_data.get()+this_pixel*num_images,
            image_memory_mean+3*this_pixel);
    }
    }
    else
    { // process ONLY litframes images in the data set    
    
    
    for (std::uint32_t this_pixel = 0; this_pixel<image_size; this_pixel+=numPixelPerTask)
    {   
        pool.submit_detach(
            process_litframes_pixel,
            transposed_data.get()+this_pixel*num_images,
            image_memory_mean+3*this_pixel); 
    }  
    }

    pool.wait_for_tasks(); 

   
}




/*
void running_stats_threadpool(
    float* data,
    std::uint32_t data_size,
    double* running_mean_mem, // of size data_size*2
    float* nanmean_of_data) // of size   1
{

    

    static_assert(2*sizeof(double)==sizeof(running_mean));

    // initialize nan mean 
    running_mean nan_mean;

    constexpr std::uint32_t data_size_per_task = 1024;

    std::vector<std::pair<std::uint64_t, uint64_t>> data;
    std::vector<double> nan_mean_per_task;
    
    auto compute_for_
    // run through data 
    for (std::uint32_t idx = 0; idx < data_size; idx++)
    {   
        const double value = static_cast<double>(data[idx]);

        // we do not process non finite values
        if(!std::isfinite(value))
        {
            continue;
        }
        
        // add to nan mean
        nan_mean.add(value);

        // add to mean 
        reinterpret_cast<running_mean*>(running_mean_mem+2*idx)->add(value);
    }

    *nanmean_of_data = nan_mean.mean;

    return;


    for (std::uint16_t bunch = 0; bunch < num_bunches; ++bunch)
    {
    // shape slow_scan_dim*fast_scan_dim
    float* bunch_panelData = panelData + bunch*slow_scan_dim*fast_scan_dim;
    
    for (std::int32_t asic_num_ss = 0; asic_num_ss < 8; ++asic_num_ss)
    {
    for (std::int32_t asic_num_fs = 0; asic_num_fs < 2; ++asic_num_fs)
    {
        pool.submit_detach(
            , bunch_panelData, asic_num_ss, asic_num_fs);
    }
    }
    pool.wait_for_tasks();
    } 
}
*/

/*

void lpd_asic_histogram_impl(
    float *panelData,                   // shape = (num_bunches, slow_scan_dim, fast_scan_dim)
    std::int32_t num_bunches,
    std::int32_t slow_scan_dim,
    std::int32_t fast_scan_dim,
    std::int32_t num_histograms,
    std::int32_t* hist_entries,
    std::uint64_t* mem_offset,
    std::uint64_t* histogram,
    std::uint64_t* blockIds)
{
    static task_thread_pool::task_thread_pool pool{16};

    if ((slow_scan_dim != 256) && (fast_scan_dim != 256))
    {   
        std::cerr << "Dimensions are off, this slow scan dim and fast scan dim are not equal to 256" << slow_scan_dim << " " << fast_scan_dim << "\n";
        return;
    }
    
    std::vector<float> multiplied{};

    for (auto count = 0; count < num_histograms; ++count)
    {
        const auto min_value = static_cast<float>(hist_entries[count*3+0]);
        const auto max_value = static_cast<float>(hist_entries[count*3+1]);
        const auto num_bins = static_cast<float>(hist_entries[count*3+2]);
        
        multiplied.emplace_back(num_bins / (max_value - min_value));
    }

    auto slow_fast_to_histogram_x = [](std::int32_t slow, std::int32_t fast) -> std::int32_t
    {
        if ((slow > 255) || (fast > 255))
        {
            std::cout << "what? : slow  fast = " << slow << " " << fast << "\n";
        }
        // slow 16x16 block index within asic
        const std::int32_t slow_block_idx = (slow / 16) % 2;
        // slow 16x16 block index within asic
        const std::int32_t fast_block_idx = (fast / 16) % 8;
        
        // get asic number
        const std::int32_t asicNum_ss = slow / 32;
        const std::int32_t asicNum_fs = fast / 128;
        const std::int32_t asicNum = asicNum_fs + asicNum_ss*2;
        
        const auto idx = asicNum*2*8 + slow_block_idx + 2*fast_block_idx;
        if (idx >= 256)
        {
            std::cout << "block num for " << slow  << " " << fast << " error , value exceeds 255, it is " << idx << "\n";
            return 0;
        }
        return idx;
    };
    
    auto get_value = [num_bunches, slow_scan_dim, fast_scan_dim](
        const float* bunch_data, 
        std::int32_t ss, 
        std::int32_t fs) -> std::optional<float>
        {
            if ((ss < 0) || (ss >= slow_scan_dim))
            {
                return std::nullopt;
            }
            if ((fs < 0) || (fs >= fast_scan_dim))
            {
                return std::nullopt;
            }

            const auto val = bunch_data[fs+ss*fast_scan_dim];
            if (!std::isfinite(val))
            {
                return std::nullopt;
            }
            return val;
        };

    auto compute_for_asic = [&](float* bunch_panelData, std::size_t asic_num_ss, std::size_t asic_num_fs)
    {
        //std::size_t counts = 0;
        //auto t1 = std::chrono::high_resolution_clock::now();

        for (std::int32_t slow = 0 + asic_num_ss*32; slow < (asic_num_ss+1)*32; ++slow)
        {
            for (std::int32_t fast = 0 + asic_num_fs*128; fast < (asic_num_fs+1)*128; ++fast)
            {
                //counts++;
        
        const auto opt_val = get_value(bunch_panelData, slow, fast);
        if (!opt_val)
        {
            continue;
        }

        const auto val = opt_val.value();
        
        const auto blockNum = slow_fast_to_histogram_x(slow, fast);
        blockIds[slow*256+fast] = blockNum;

        for (std::size_t hist_number = 0; hist_number < static_cast<std::size_t>(num_histograms); ++hist_number)
        {
            const auto min_value = static_cast<float>(hist_entries[hist_number*3+0]);
            const auto max_value = static_cast<float>(hist_entries[hist_number*3+1]);
            const auto num_bins = static_cast<float>(hist_entries[hist_number*3+2]);
            if ((val < min_value) || (val >= max_value))
            {
                continue;
            }

            const auto binNr  = static_cast<std::size_t>(std::floor(multiplied[hist_number]*(val-min_value)));
            histogram[static_cast<std::size_t>(mem_offset[hist_number] + blockNum*num_bins+binNr)] += 1;

        }

            }
        }
        
        //auto t2 = std::chrono::high_resolution_clock::now();
        //auto us_hist = static_cast<std::int64_t>(std::chrono::duration_cast<std::chrono::microseconds>(t2 - t1).count());
        //std::cout << "time to add for asic:" << us_hist << "\n";
        
    };
    
    // execution model: one task for each 128x32 asic, 
    
    for (std::uint16_t bunch = 0; bunch < num_bunches; ++bunch)
    {
    // shape slow_scan_dim*fast_scan_dim
    float* bunch_panelData = panelData + bunch*slow_scan_dim*fast_scan_dim;
    
    for (std::int32_t asic_num_ss = 0; asic_num_ss < 8; ++asic_num_ss)
    {
    for (std::int32_t asic_num_fs = 0; asic_num_fs < 2; ++asic_num_fs)
    {
        pool.submit_detach(
            compute_for_asic, bunch_panelData, asic_num_ss, asic_num_fs);
    }
    }
    pool.wait_for_tasks();
    }    
}
*/