

#import matplotlib
#import matplotlib.pyplot as plt
#from matplotlib.backends.backend_agg import FigureCanvasAgg


#import os
import ctypes
import numpy as np
import time

from multiprocessing import Process, Pipe, Array, Value

def tonumpyarray(shared_mem_array):
    return np.frombuffer(shared_mem_array.get_obj(), dtype=np.uint8)

def plot_receiver(connection, shared_mem_for_plot, shared_mem_trainId):
    '''
    this function makes the plot for the single shot contrast'''
    print('Receiver: Running', flush=True)
    # consume work

    
    while True:
        # get a unit of work
        item = connection.recv()

        if item is None:
            break
        
        output = item['fun'](*item['arg']) 
        npArray = tonumpyarray(shared_mem_for_plot)
        npArray[:] = output.reshape(-1)
        shared_mem_trainId.value = item['train_id']


        
    # all done
    print('Receiver: Done', flush=True)


class OnlineImagePlotter():

    def __init__(self, figsize = (15,10)):
        self.figsize = figsize
        arraysize_bytes = (self.figsize[0]*100)*(self.figsize[1]*100)*4

        self.plot_memory = Array(ctypes.c_ubyte, [0]*arraysize_bytes)
        self.train_id =  Value(ctypes.c_ulonglong, 0)
        self.prev_train_id = -1

        #pipe
        receiver, sender = Pipe()
        self.receiver = receiver
        self.sender = sender

        self.process = Process(target=plot_receiver, args=(self.receiver, self.plot_memory, self.train_id), daemon=True)
        self.process.start()

    def close(self):
        self.sender.send(None)
        self.process.join(3)
        print("is_alive() after join?", self.process.is_alive())
        self.process.close()
        print("is_alive() after close?", self.process.is_alive())
        self.process.kill()
        print("is_alive() after kill?", self.process.is_alive())
        

    def get_image_buffer_if_ready_and_make_new_image(self, train_id, fun, *args):
        '''
        return None if image is not ready
        return np array of shape (fihsize[1]*100, figsize[1]*100, 4) otherwise
        which then can be used in the output hash
        -> output_hash[' hash name '] = ImageData(np array from this function)
        '''

        current_image = None

        def is_new_image_ready():
            return self.prev_train_id != self.train_id.value

        if is_new_image_ready():
            
            # get the previously computed image

            def get_plot_as_np_array(shared_mem_array):
                np_array = tonumpyarray(shared_mem_array)
                return np.reshape(np_array, (self.figsize[1]*100, self.figsize[0]*100, 4))
            
            current_image = get_plot_as_np_array(self.plot_memory)

            self.prev_train_id = self.train_id.value

            # ask for next image to be made
            package = dict()
            package['train_id'] = train_id
            package['fun'] = fun
            package['arg'] = args
            self.sender.send(package)
        else:
            pass
        return current_image


class OnlinePlotHandler():

    def __init__(self):

        self.keys_and_plotter = dict()

    def is_key_registered(self, key):
        if key not in self.keys_and_plotter:
            print(f"key {key} does not exist, call 'register' function in initialization() part of this karabo device!")
            return False
        return True
        
    def register(self, key, Plotter):
        self.keys_and_plotter[key] = { 
            "plotter" : Plotter(),
            "time_to_make_plot" : time.time()}
        
    
    
    def add_schema(self, key, schema):
        if not self.is_key_registered(key):
            return schema
        
        from karabo.bound import (
            NODE_ELEMENT,
            IMAGEDATA_ELEMENT,
            OUTPUT_CHANNEL,
            STRING_ELEMENT,
            Schema
            )
        
            
        
        (
            NODE_ELEMENT(schema)
            .key(key)
            .displayedName(f'{key}')
            .commit(),

            IMAGEDATA_ELEMENT(schema)
            .key(f"{key}.plot")
            .displayedName('plot')
            .description('plot')
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{key}.update_interval")
            .displayedName('Update interval')
            .description('Update interval of the plot')
            .assignmentOptional()
            .defaultValue("")
            .commit(),
        )
        
        return schema
    
    def close(self):
        for key, item in self.keys_and_plotter.items():
            print(f"online plot handler: closing {key}")
            item['plotter'].close()
            
    

    def save_to_file(self, filename, current_img):
        pass

    def save(self, hash=None, filename=None, key=None, plotter_args=None):
        if key is None or plotter_args is None:
            print("Error in OnlinePlotHandler:save() function, key and args must be given ")

        if not self.is_key_registered(key):
            if hash is not None:
                return hash
            return
        
        current_img = self.keys_and_plotter[key]['plotter'].plot(*plotter_args)

        if current_img is None:
            if hash is not None:
                return hash
            return    
        
        if filename is not None:
            self.save_to_file(filename, current_img)
        
        if hash is not None:
            from karabo.bound import ImageData
            
            hash[f'{key}.plot'] = ImageData(current_img)

            now = time.time()
            hash[f'{key}.update_interval'] = f'{(now - self.keys_and_plotter[key]['time_to_make_plot']):.1f}s' 
            self.keys_and_plotter[key]['time_to_make_plot'] = now
        
            return hash
    
    
